from sklearn.model_selection import train_test_split as split
from imblearn.over_sampling import SMOTE
from multiprocessing import Pool
from scipy.stats import sem
import numpy as np


class CV():
    def __init__(self, model, metric, lower_bound=False):
        self.model, self.metric = model, metric
        self.lower_bound = lower_bound
        
        return None
    
    def cv_unitary(self, iterator):
        metric_list = []
        for i in range(self.cv):
            x_train, x_test, y_train, y_test = split(self.x, self.y, test_size=self.frac, stratify=self.y)
            
            if self.over and self.workers == 1:
                oversample = SMOTE(n_jobs=-1)
                x_train, y_train = oversample.fit_resample(x_train, y_train)
            
            else:
                oversample = SMOTE()
                x_train, y_train = oversample.fit_resample(x_train, y_train)
    
            self.model.fit(x_train, y_train)
            predictions = self.model.predict(x_test)
            
            score = self.metric(y_test, predictions)
            metric_list.append(score)
            
        if self.full:
            return metric_list
        
        elif self.lower_bound:
            return np.mean(metric_list) - 1.96*sem(metric_list)
        
        else:
            return np.mean(metric_list)
        
    def cv_parallel(self):
        with Pool(self.workers) as pool:
            metric_list = pool.map(self.cv_unitary, range(self.solutions))
            pool.close()
            pool.join()
            
        if self.full:
            return metric_list
        
        elif self.lower_bound:
            return np.mean(metric_list) - 1.96*sem(metric_list)
        
        else:
            return np.mean(metric_list)
        
    
    def cv(self, x, y, cv=20, frac=0.1, over=True, full=False, workers=1):
        self.x, self.y = x, y
        self.frac = frac
        self.over, self.full = over, full
        self.workers=workers
        
        if workers == 1:
            self.cv = cv
            out = self.cv_unitary(iterator=0)
            
        elif workers > 1:
            self.solutions = cv
            self.cv = 1
            
            out = self.cv_parallel()
            
        else:
            return 'Number of Workers must be >= 1'
            
        del [self.x, self.y, self.cv, self.frac, self.over, self.full, self.workers]
            
        return out


from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import f1_score
from multiprocessing import Pool
import pandas as pd
import numpy as np
import warnings

warnings.filterwarnings('error', category=ConvergenceWarning)

##################
# Model Function #
##################
def model_fit(x_train, y_train, x_test, y_test, model, pop, i):
    if model == 'logit':
        solver = 'lbfgs' if 'solver' not in pop else pop['solver'][i]
        penalty = 'l2' if 'penalty' not in pop else pop['penalty'][i]
        
        model = LogisticRegression(penalty=penalty, solver=solver)
        
        try:
            model.fit(x_train, y_train)
            predictions = model.predict(x_test)
            return f1_score(y_test, predictions, average='micro')
        
        except ConvergenceWarning:
            return 0
    
    elif model == 'rfc':
        n_samples = 1000 if 'n_samples' not in pop else pop['n_samples'][i]
        max_depth = None if 'max_depth_rfc' not in pop else pop['max_depth_rfc'][i]
        
        model = RandomForestClassifier(n_samples=n_samples, max_depth=max_depth)
        model.fit(x_train, y_train)
        predictions = model.predict(x_test)
        return f1_score(y_test, predictions, average='micro')
    
    elif model == 'gbc':
        learning_rate = 0.1 if 'learning_rate' not in pop else pop['learning_rate'][i]
        n_estimators = 100 if 'n_estimators' not in pop else pop['n_estimators'][i]
        max_depth = 3 if 'max_depth_gbc' not in pop else pop['max_depth_gbc'][i]
        
        model = GradientBoostingClassifier(n_estimators=n_estimators, learning_rate=learning_rate, max_depth=max_depth)
        model.fit(x_train, y_train)
        predictions = model.predict(x_test)
        return f1_score(y_test, predictions, average='micro')


class GA:
    def __init__(self, x, y, pop, cv=[], model='logit', max_features=None):
        self.x, self.y = x, y
        self.pop, self.cv = pop, cv
        self.max_features = max_features
        self.model = model
        self.max = []

    ####################
    # Fitness Function #
    ####################
    def fit(self, i=0):
        max_features = self.max_features if self.max_features != None else self.x.shape[1]
        rows = self.x.loc[self.pop['rows'][i] == 1].index
        cols = self.pop['cols'][i]
        
        fit_list = []
        
        if sum(cols) > max_features:
            fit_list.append(0)
        
        else:
            x = self.x.dropna()
            y = self.y.loc[x.index]
            
            x = x[[x.columns[j] for j in range(x.shape[1]) if cols[j] == 1]]      
            
            for fold in self.cv:
                x_test = x.loc[x.index.isin(fold)]
                
                x_train = x.loc[~x.index.isin(fold)]
                x_train = x_train.loc[~x_train.index.isin(rows)]
                
                y_test = y.loc[x_test.index]
                y_train = y.loc[x_train.index]
                
                fit_list.append(model_fit(x_train, y_train, x_test, y_test, self.model, self.pop, i))
        
        return np.mean(fit_list)
    
    ###################
    # Mating Function #
    ###################
    def mate(self, fitness, num_parents):
        return self.pop.loc[self.pop.index.isin(((-np.array(fitness)).argsort())[0:num_parents])]
    
    #####################
    # Children Function #
    #####################
    def children(self, fitness, num_children, mutation_rate):
        children = pd.DataFrame(columns=self.pop.columns, index=range(num_children))
        
        for i in range(num_children):
            mates = np.random.choice(np.arange(0, len(self.pop)), size=2, p=fitness/sum(fitness))
            mom, dad = self.pop.iloc[mates[0]], self.pop.iloc[mates[1]]
            
            for j in range(len(self.pop.columns)):
                col = self.pop.columns[j]
                if type(mom[col]) == int:
                    mutate = bool(np.random.binomial(1, mutation_rate))
                    children.iloc[i, j] = round(np.mean([mom[col], dad[col]])) if mutate else np.random.choice([mom[col], dad[col]])
                    
                elif type(mom[col]) == float:
                    mutate = bool(np.random.binomial(1, mutation_rate))
                    children.iloc[i, j] = np.mean([mom[col], dad[col]]) if mutate else np.random.choice([mom[col], dad[col]])
                    
                elif type(mom[col]) == str:
                    mutate = bool(np.random.binomial(1, mutation_rate))
                    children.iloc[i, j] = str(np.random.choice(list(set(self.pop[col]))) if mutate else np.random.choice([mom[col], dad[col]]))
                
                elif type(mom[col]) == np.ndarray:
                    mutate = np.random.binomial(1, mutation_rate, len(mom[col]))
                    choice = np.random.binomial(1, 0.5, len(mom[col]))
                    child = [mom[col][i] if choice[i] == 1 else dad[col][i] for i in range(len(choice))]
                    children.iloc[i, j] = np.array([1 - child[i] if mutate[i] == 1 else child[i] for i in range(len(mutate))])
                
                else:
                    print('Invalid Data Type')
                    print(f'{mom[col]} needs to be data type int, float, str, or ndarray')
                    print(f'{mom[col]} is currently data type {type(mom[col])}')
                    print('')
                    break
                    
        return children
        
    #######################
    # Parallelize Fitness #
    #######################
    def parallel(self, solutions, workers=2):
        pool = Pool(workers)
        out = pool.map(self.fit, range(solutions))
        pool.close()
        pool.join()
        
        return out
    
    
        
    
    


from collections import OrderedDict
from torch import nn
import numpy as np
import torch

class Discriminator():
    def __init__(self, drop=0.1, neurons=[64, 32, 16], lr_nn=0.0001, 
                 epochs=20, layers=3, device='cuda:0', batch_size=2**7):
        self.drop, self.layers = drop, layers
        self.neurons = neurons
        self.lr_nn, self.epochs = lr_nn, epochs
        self.device = torch.device(device)
        self.batch_size = batch_size
        
        return None
        
    class Classifier(nn.Module):
        def __init__(self, shape, neurons, drop, output, layers=3):
            super().__init__()
            
            neurons = [shape] + neurons
            sequential = OrderedDict()
            
            i = 0
            while i < layers:
                sequential[f'linear_{i}'] = nn.Linear(neurons[i], neurons[i+1])
                sequential[f'relu_{i}'] = nn.ReLU()
                sequential[f'drop_{i}'] = nn.Dropout(drop)
                i+=1
                
            sequential['linear_final'] = nn.Linear(neurons[i], output)
            sequential['softmax'] = nn.Softmax(dim=1)
            
            self.model = nn.Sequential(sequential)
            
        def forward(self, x):
            out = self.model(x)
            return out
    
    def fit(self, x, y):
        col_count = x.shape[1]
        output = len(set(y))
        
        x, y = torch.from_numpy(x.values).to(self.device), torch.from_numpy(y.values).to(self.device)
        
        train_set = [(x[i].to(self.device), y[i].to(self.device)) for i in range(len(y))]
        train_loader = torch.utils.data.DataLoader(train_set, batch_size=self.batch_size, shuffle=True)
    
        loss_function = nn.CrossEntropyLoss()
        discriminator = self.Classifier(col_count, self.neurons, self.drop, output, self.layers).to(self.device)
        optim = torch.optim.Adam(discriminator.parameters(), lr=self.lr_nn)
    
        for epoch in range(self.epochs):
            for i, (inputs, targets) in enumerate(train_loader):
                discriminator.zero_grad()
                yhat = discriminator(inputs.float())
                loss = loss_function(yhat, targets.long())
                loss.backward()
                optim.step()
                
        self.model = discriminator
        
        return None
    
    def predict(self, x):
        discriminator = self.model
        discriminator.to(self.device).eval()
        
        x = torch.from_numpy(x.values).to(self.device)
        preds = np.argmax(discriminator(x.float()).cpu().detach(), axis=1)
        
        return preds
    
    def predict_proba(self, x):
        discriminator = self.model
        discriminator.to(self.device).eval()
        
        x = torch.from_numpy(x.values).to(self.device)
        preds = discriminator(x.float()).cpu().detach()
        
        return preds



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
      name='zangorth-helpers',
      version='1.3.3',
      description='Collection of helper functions for my projects',
      long_description=README,
      long_description_content_type='text/markdown',
      url='https://github.com/Zangorth/Helpers',
      author='Zangorth',
      packages=['helpers'],
      install_requires=open(r'C:\Users\Samuel\Google Drive\Portfolio\Helpers\requirements.txt', 'r').read().split('\n')[0:-1]
      )